#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
===============================================================================
Created on    : Fri Jun 25 21:00:10 2021
Author        : Alejandro Martínez León
Mail          : [alejandro.martinezleon@uni-saarland.de, ale94mleon@gmail.com]
Affiliation   : Jochen Hub's Biophysics Group
Affiliation   : Faculty of NS, University of Saarland, Saarbrücken, Germany
===============================================================================
DESCRIPTION   :
DEPENDENCIES  :
===============================================================================
"""
import os, subprocess, shutil, tempfile, time, inspect
import numpy as np
import multiprocessing as mp
from glob import glob




def timeit(method):
    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        if 'log_time' in kw:
            name = kw.get('log_name', method.__name__.upper())
            kw['log_time'][name] = int((te - ts) * 1000)
        else:
            print('%r  %2.2f ms' % \
                  (method.__name__, (te - ts) * 1000))
        return result
    return timed

        

def run(command, shell = True, executable = '/bin/bash', Popen = False):
    #Here I could make some modification in order that detect the operator system
    #NAd make the command compatible with the opertor system
    #the fucntion eval could be an option if some modifcation to the variable command 
    #need to be done.... SOme fligth ideas...

    if Popen:
        #In this case you could acces the pid as: run.pid
        process = subprocess.Popen(command, shell = shell, executable = executable)
    else:
        process = subprocess.run(command, shell = shell, executable = executable)
    return process

@timeit
def mopac(mop):
    print(f"Mopac is running ...")
    run(f"mopac {mop}  > /dev/null 2>&1")
    print("Done!")


def job_launch(shell="sbatch", script_name = "job.sh"):
    """
    

    Parameters
    ----------
    shell : TYPE, optional
        DESCRIPTION. The default is "sbatch".
    script_name : TYPE, optional
        DESCRIPTION. The default is "job.sh".
        If a regular expresion is provided
        then the function will execute the first ordered alphabetically. E.g:
            job.* was provided and there job.sh and job.bash. Then it will use job.bash.


    Returns
    -------
    JOBIDs : list of integers.
        DESCRIPTION. The ID of the launch in case of sbatch was used as shell

    """
    cwd = os.getcwd()
    JOBIDs = []
    for (root, dirs, files) in list(os.walk(cwd)):
        os.chdir(root)
        if glob(script_name):
            script_name = sorted(glob(script_name))[0]
            JOBID_tmp_file = tempfile.NamedTemporaryFile()
            if shell == "sbatch":
                run(f"{shell} --parsable {script_name}>{JOBID_tmp_file.name}")
                with open(JOBID_tmp_file.name, "r") as f:
                    JOBIDs.append(int(f.readline()))
            else:
                run(f"{shell} {script_name}>{JOBID_tmp_file.name}")
                try:
                    with open(JOBID_tmp_file.name, "r") as f:
                        JOBIDs.append(int(f.readline()))
                except:
                    pass

    os.chdir(cwd)
    return JOBIDs


#=======================================================================================

#                          Tools for working with files

#=======================================================================================
def makedirs(path):
    os.makedirs(path,exist_ok=True)

def rm(pattern, r = False):
    """
    

    Parameters
    ----------
    patterns : string
        input-like Unix rm
    r : TYPE, bool
        DESCRIPTION. The default is False.
        If True delete also directories
    Returns
    -------
    None.

    """
    if glob(pattern):
        for p in glob(pattern):

            if r:
                if os.path.isdir(p):
                    shutil.rmtree(p)
                elif os.path.isfile(p):
                    os.remove(p)       
            else:
                if os.path.isfile(p):
                    os.remove(p)
                elif os.path.isdir(p):
                    print(f"The directory '{p}' was not deleted, set the flag r to True")
    else:
        
        print(f"rm: '{pattern}' doesn't exist")

                
def cp(src, dest, r = False):
    """
    
    This function makes use of the possible multiple CPU of the machine.
    Parameters
    ----------
    src : TYPE: string
        DESCRIPTION:
            Source Path to a directory or a file. regular expresion are accepted. The librery glob is used for that.
    dest : TYPE: string
        DESCRIPTION:
            Destination Path
    r : TYPE, optional
        DESCRIPTION. The default is False:
            If True, Also directories will be copy. If not, and a directory was
            given as src, a Raise Exception will be printed
            Another Raise Exception will be printed if the destination path doesn't
            exist.

    Returns
    -------
    None.

    """
    src = os.path.abspath(src)
    dest = os.path.abspath(dest)
    #This scheme doesn't consider the possibilities of other item except files, dir or regualr expresions
    if glob(src) and os.path.exists(os.path.dirname(dest)):
    
        if os.path.isdir(src):
            if r == True:            
                
                basename_src = os.path.basename(src)
                for root, dirs, files in os.walk(src):
                    path2copy = root.split(src)[1]
                    try:
                        if path2copy[0] == "/":
                            path2copy = path2copy[1:]
                    except: pass
                    path_dest = os.path.join(dest,basename_src,path2copy)
                    
                    if dirs:
                        pool = mp.Pool(mp.cpu_count())
                        pool.map(makedirs, [os.path.join(path_dest, d) for d in dirs])
                        pool.close()
                    if files:
                        makedirs(path_dest)
                        pool = mp.Pool(mp.cpu_count())
                        
                        pool.starmap(shutil.copy2, [(os.path.join(root, f),path_dest) for f in files])
                        pool.close()
            else:
                print(f"If you need to copy directories, set the 'r' flag to True. {src} is a directory")
    
        elif os.path.isfile(src):
            shutil.copy2(src,dest)
            #if not os.path.exists(dest):
             #   print(f"The file '{src} was not copy to '{dest}' becasue doesn't exist or is not accesible")
        
        
        else:#Here we are in regular expresions
            to_copy = [file for file in glob(src) if os.path.isfile(file)]
            pool = mp.Pool(mp.cpu_count())
            pool.starmap(shutil.copy2, [(pattern,dest) for pattern in to_copy])
            pool.close()
        
    elif not glob(src) and not os.path.exists(os.path.dirname(dest)):
        print(f"cp: neither {src}, nor {dest} exist.")
    elif not glob(src):
        print(f"cp: The source file {src} doesn't exist")
    else:
        print(f"cp: cannot create regular file '{dest}': Not a directory")
            
        

def mv(src, dest, r = False):
    cp(src, dest, r = r)
    rm(src,r=r)

def KbT(absolute_temperature):
    """Return the value of Kb*T in kJ/mol

    Args:
        absolute_temperature (float): The absolute temperature in kelvin
    """
    Kb = 8.314462618E-3 #kJ/(mol⋅K) (kNA)
    return absolute_temperature*Kb
def get_default_args(func):
    signature = inspect.signature(func)
    return {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }

if __name__ == '__main__':  
    p = get_default_args(run)
    print(type(p['shell']))

  


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from aleimi import confgen, boltzmann, extractor, tools, __version__
import argparse
import os

"""
Tengo que adicionar la parte de los comandos extras para pasarselos a exrtractor
para la creacion de los templates, probar con los argumentos que se pasan
si alguno esta mal entonces dlanzar un warning o algo por el estilo
Tengo que ver esto bien, me falta por implemnetar lo fde .gjf
Y lo de el analisis para al menos psi4 y orca
    """

def _aleimi():
    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument(
        help = "The path were the molecule(s) is(are)",
        dest='suppl',
        type=str)
    parser.add_argument(
        '-p', '--params',
        help = "Parameters to run ALEIMI",
        dest='params',
        type=str)
    parser.add_argument(
        '-v', '--version',
        action='version',
        version=f"aleimi: {__version__}")

    args = parser.parse_args()

    suppl = args.suppl
    params = args.params
    if not os.path.exists(suppl):
        raise FileNotFoundError(f"{suppl} does not exist or is not accessible.")
    if not os.path.exists(params):
        raise FileNotFoundError(f"{params} does not exist or is not accessible.")



    with open(params,'r') as f:
        lines = f.readlines()

    user_keywords = {}
    for line in lines:
        if not line.startswith('#') and len(line.strip()):
            line = line.split('#')[0]
            index = line.index('=') # Only get the first equal
            key, value = line[:index].replace('-', '_').strip(), line[index+1:].strip()
            user_keywords[key] = value
    user_keys = user_keywords.keys()

    confgen_keywords = tools.get_default_args(confgen.main)
    for key in confgen_keywords.keys():
        if key in user_keys:
            try:
                confgen_keywords[key] = type(confgen_keywords[key])(user_keywords[key]) # Here I  am taking the type of the variable
            except:
                raise ValueError(f"{user_keywords[key]} must be a {type(confgen_keywords[key])}-like")

    mopac_keywords = tools.get_default_args(tools.mopac)
    for key in mopac_keywords.keys():
        if key in user_keys:
            try:
                mopac_keywords[key] = type(mopac_keywords[key])(user_keywords[key]) # Here I  am taking the type of the variable
            except:
                raise ValueError(f"{user_keywords[key]} must be a {type(mopac_keywords[key])}-like")

    boltzmann_keywords = tools.get_default_args(boltzmann.main)
    for key in boltzmann_keywords.keys():
        if key in user_keys:
            try:
                boltzmann_keywords[key] = type(boltzmann_keywords[key])(user_keywords[key]) # Here I  am taking the type of the variable
            except:
                raise ValueError(f"{user_keywords[key]} must be a {type(boltzmann_keywords[key])}-like")

    extractor_keywords = tools.get_default_args(extractor.main)
    for key in extractor_keywords.keys():
        if key in user_keys:
            try:
                extractor_keywords[key] = type(extractor_keywords[key])(user_keywords[key]) # Here I  am taking the type of the variable
            except:
                raise ValueError(f"{user_keywords[key]} must be a {type(extractor_keywords[key])}-like")

    with open('outparams.params', 'w') as f:
        used_keywords = {**confgen_keywords, **mopac_keywords, **boltzmann_keywords, **boltzmann_keywords}
        for key in used_keywords:
            f.write(f"{key:<35} = {used_keywords[key]:<20}\n")

    mol_names = confgen.main(suppl,**confgen_keywords)
    for mol_name in mol_names:
        tools.mopac(f"{mol_name}.mop",**mopac_keywords)
        boltzmann.main(f"{mol_name}.arc",**boltzmann_keywords)
        extractor.main(f"{mol_name}.arc",f"{mol_name}.boltzmann", **boltzmann_keywords)


# -*- coding: utf-8 -*-
"""
Toma de un archivo .smi los SMILES, genera la cant de conformeros seleccionados
e imprime tanto los .sdf con todos los conformeros como .mop para optimizar en MOPAC
con el nivel de teoria seleccionado.
"""

from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem import Draw
import glob as glob
import pandas as pd
# from operator import itemgetter
import os
import tempfile
from aleimi import OBconvert, tools

#      CHECKING geometry degeneracy
# La primera iteracion que corre por i busca que este optimizada la estructura
# La segunda busca que no se geometricamente la misma estructura
#Pueda suceder que alguna idx cumpla que no es geometricamente la misma estructura
#y no la ve el segundo ciclo
#pero el primero cuando llegue a ella la vera y tomara su id
# =============================================================================
def confgen(mol, rdkit_d_RMSD, numConfs, rdkit_numThreads = 0, UFF = False):
    molecule = Chem.AddHs(mol)
    ids = AllChem.EmbedMultipleConfs(molecule, numConfs=numConfs, numThreads=rdkit_numThreads)
    if UFF:
        opt = AllChem.UFFOptimizeMoleculeConfs(molecule, numThreads=rdkit_numThreads, ignoreInterfragInteractions=False, maxIters=500) # The result is a list a containing 2-tuples: (not_converged, energy) for each conformer. If not_converged is 0, the minimization for that conformer converged.

        print('Calculando RMSD ...')
        rmsd_reject = set()
        opt_reject = set()
        rmsd_no_reject= set()
        no_opt=0
        ids=list(ids)
        for i in ids:

            if opt[i][0]==1:
                no_opt +=1
                opt_reject.add(ids[i])
            else:
                for idx in ids:
                    if i < idx:
                        RMSD = AllChem.GetConformerRMS(molecule, i, idx, prealigned=True)
                        if RMSD <= rdkit_d_RMSD:
                            if opt[i][1] < opt[idx][1]:
                                rmsd_reject.add(ids[i])
                                rmsd_no_reject.add(ids[idx])
                            else:
                                rmsd_reject.add(ids[idx])
                                rmsd_no_reject.add(ids[i])

        reject = opt_reject.union(rmsd_reject)
        if len(rmsd_reject) !=0:
            print(f'Los confórmeros con Id {rmsd_reject} no se imprimirán en el .mop por ser iguales (Según el error de RMSD = {rdkit_d_RMSD}) a {rmsd_no_reject} respectivamente y menos favorables energéticamente que los últimos mencionados.')
        if len(opt_reject) !=0:
            print(f'Los confórmeros con Id {opt_reject} no se imprimirán en el .mop pues no lograron ser optimizados.')
        to_use = list(set(ids) - reject)
        return molecule, to_use, [item[1] for item in opt]# Here I am creating a list of tuple (molecule, energy)
    else:

        print('Calculando RMSD ...')
        rmsd_reject = set()
        rmsd_no_reject= set()
        ids=list(ids)
        for i in ids:
            for idx in ids:
                if i < idx:
                    RMSD = AllChem.GetConformerRMS(molecule, i, idx, prealigned=True)
                    if RMSD <= rdkit_d_RMSD:
                        rmsd_reject.add(ids[i])
                        rmsd_no_reject.add(ids[idx])


        if len(rmsd_reject) !=0:
            print(f'Los confórmeros con Id {rmsd_reject} no se imprimirán en el .mop por ser iguales (Según el error de RMSD = {rdkit_d_RMSD}) a {rmsd_no_reject} respectivamente y menos favorables energéticamente que los últimos mencionados.')

        to_use = list(set(ids) - rmsd_reject)
        return molecule, to_use, len(ids)*[None] # Because I did not optimize, then I add None

# =============================================================================
#exportar las imagenes de la mejor manera posible. en columnas de 1,2,3,4 o 5
# =============================================================================

def makeimg(mols, **keywords):

    ms = [mol for mol in mols if mol is not None]
    [AllChem.Compute2DCoords(m) for m in ms]

    cant = len(ms)
    if cant==1:
        best=1
    else:
        div = range(2,6)
        tup = [[i - cant%i,i] for i in div]

        for t in tup:
            if t[0] == t[1]:
                t[0] = 0

        minimum_rest = min(tup, key=lambda x: x[0])
        for t in tup:
            if t[0] == minimum_rest[0] and t[1]>minimum_rest[1]:
                minimum_rest = t
        best = minimum_rest[1]

    if 'legends' in keywords:
        legends = keywords['legends']
    else:
        legends = [f'Mol: {x+1}' for x in range(len(ms))]

    img=Draw.MolsToGridImage(ms,molsPerRow=best,subImgSize=(700,700), legends=legends)
    if len(ms) > 1:
        img.save('Molecules used.png')
    else:
        img.save(f'{legends[0]}.png')




def main(suppl, numConfs = 10, rdkit_d_RMSD = 0.2, UFF = False, rdkit_numThreads = 0, mopac_keywords =  'PM7 precise ef xyz geo-ok t=3h THREADS = 2'): #EPS=78.4

    ext = os.path.basename(suppl).split('.')[-1]
    name = os.path.basename(suppl)[:-(len(ext)+1)]
    print('Resumen\n')
    if ext == "smi":
        with open(suppl, 'rt') as file:
            smiles = file.readlines()

        # =============================================================================
        #    Comprobamos que los SMILES esten bien
        # expuestas con la funcion molfilter y exportamos .sdf
        # =============================================================================
        Errors_mols = []
        for i, item in enumerate(smiles):
            try:
                posibble_molecule = (Chem.MolFromSmiles(item))
                posibble_molecule.GetNumAtoms()
            except:
                Errors_mols.append(i+1)
        if len(Errors_mols) !=  0:
            raise ValueError(f'Las moléculas con Ids: {Errors_mols} no son estructuras SMILES correctas para RDKit. Por favor, retirelas del .smi.')




        mols = [(f'conf_mol_{i+1}', Chem.MolFromSmiles(smile)) for (i,smile) in enumerate(smiles)]
    elif ext == 'pdb':
        mols = [(f"conf_{name}", Chem.MolFromPDBFile(suppl))]
    elif ext == 'mol2':
        mols = [(f"conf_{name}", Chem.MolFromMol2File(suppl))]
    elif ext == 'mol':
        mols = [(f"conf_{name}", Chem.MolFromMolFile(suppl))]
    elif ext == 'mol':
        mols = [(f"conf_{name}", Chem.MolFromMolFile(suppl))]
    else:
        print('The molecule will be internally converted to .mol using openbabel')
        tmpfile = tempfile.NamedTemporaryFile(suffix='.mol')
        OBconvert.obconvert(suppl, tmpfile.name)
        mols = [(f"conf_{name}", Chem.MolFromMolFile(tmpfile.name))]
        tools.rm(tmpfile.name)

    if None in [mol[1] for mol in mols]:
        raise ValueError(f"{suppl} is not understand by neither RDKit nor OpenBabel")




    # =============================================================================
    #      Generamos conformaciones que cumplen con las condiciones de rmsd
    # expuestas con la funcion molfilter y exportamos .sdf
    # =============================================================================

    makeimg([mol[1] for mol in mols], legends = [mol[0] for mol in mols])
    for (i, mol) in enumerate(mols):

        print(f'Se están generando {numConfs} conformaciones para la molécula con Id = {i+1}...')

        molecule, index, opt = confgen(mol[1], rdkit_d_RMSD, numConfs, rdkit_numThreads = rdkit_numThreads, UFF = UFF)
        natoms = molecule.GetNumAtoms()

        sdf_writer = Chem.SDWriter(f"{mol[0]}.sdf")
        for idx in index:
            sdf_writer.write(molecule, confId=idx)
        sdf_writer.close()
        print(f'{mol[0]}:  {len(index)}/{numConfs}')

    # =============================================================================
    #     Generando los .mop
    # =============================================================================


        print(f"Archivos de salida: {mol[0]}.sdf, {mol[0]}.mop")

        with open(f"{mol[0]}.sdf", 'rt') as file:
            lines = file.readlines()

        final=open(f"{mol[0]}.mop",'w')

        #Toma los valores reales que se usaron para imprimir el .sdf y asi buscar los opt de verdad, pq si se elimino una estructura tengo que tenerlo en cuenta
        cont = 0 #Lo inicializao en -1 pq la cantidad de veces que se encontrara la palabra RDKit coincide con la cantidad con el len(to_use[0])
        #todo esto para que en el .mop me saque el valor de energia optenido en la optimizacion
        for k, line in enumerate(lines):
            if 'RDKit          3D' in line:

                chunk = lines[(k+3):(k+3+(natoms))]
                sliced = []
                for c in chunk:
                    sliced.append(c.split())
                df = pd.DataFrame(sliced)
                to_print = df[[3, 0, 1, 2]]
                final.write(mopac_keywords+'\n')
                try:
                    comments = f"{mol[0]}   E_UFF = {round(float(opt[index[cont]]), 3)}"
                except:
                    comments = f"{mol[0]}   E_UFF = {opt[index[cont]]}"
                final.write(comments+'\n')
                final.write('CELL: %d\n' % (cont+1))
                to_print.to_string(final, header=False, index=False)
                final.write('\n0\n')
                cont += 1
        print('Número de átomos: %d\n'% (natoms))
        final.close()
    return [mol[0] for mol in mols]





# -*- coding: utf-8 -*-
"""
===============================================================================
Created on    : Tue Dec  8 00:51:27 2020
Author        : Alejandro Martínez León
Mail          : [alejandro.marrtinezleon@uni-saarland.de, ale94mleon@gmail.com]
Affiliation   : Jochen Hub's Biophysics Group
Affiliation   : Faculty of NS, University of Saarland, Saarbrücken, Germany
===============================================================================
DESCRIPTION   :
DEPENDENCIES  :
===============================================================================
"""

import numpy as np
import math as mt
from rmsd import kabsch_rmsd
import math as mt
def calc_dihedral(p1, p2, p3, p4):
    """ Calculate dihedral angle method.
    The angle is in [-pi, pi].
    """
    #p1 = np.array(p1)
    #p2 = np.array(p2)
    #p3 = np.array(p3)
    #p4 = np.array(p4)

    b1 = p2 - p1
    b2 = p3 - p2
    b3 = p4 - p3
    
    n1 = np.cross(b1,b2)
    n1 = n1/np.linalg.norm(n1)
    
    n2 = np.cross(b2,b3)
    n2 = n2/np.linalg.norm(n2)
    
    b2 = b2/np.linalg.norm(b2)
    
    m1 = np.cross(n1,b2)
    
    x = (n2*n1).sum()
    y = (n2*m1).sum()
    
    return mt.atan2(y,x)

def calc_angle_bond(p1, p2, p3):
    """ Calculate bond angle method.
    The angle is in [-pi, pi].
    """
    #p1 = np.array(p1)
    #p2 = np.array(p2)
    #p3 = np.array(p3)    
    b1 = p1 - p2
    b2 = p3 - p2
    b1b2 = (b1*b2).sum()
    norm_b1 = np.linalg.norm(b1)
    norm_b2 = np.linalg.norm(b2)
    return mt.acos(b1b2/(norm_b1*norm_b2))

def calc_bond_dist(p1, p2):
    #p1 = np.array(p1)
    #p2 = np.array(p2)    
    return np.linalg.norm(p1 - p2)

def psi4out2coords(out, element = False):
    '''
    

    Parameters
    ----------
    out : string
        name of the psi4.out file from opt.
    element : TYPE, optional
        DESCRIPTION. The default is False.
        if False:
            return only the xyz coords
        if True:
            Also return the element
    Returns
    -------
    None.

    '''
    with open(out, 'rt', encoding='latin-1') as file:
        lines = file.readlines()
   
    xyz = []

    for i in range(len(lines)):
        if 'Final optimized geometry and variables:' in lines[i]:
            for j in range(i+6,len(lines)):
                if '\n' == lines[j]: break
                split = lines[j].split()    
                if element:
                    xyz.append(np.array([split[0], float(split[1]), float(split[2]), float(split[3])]))
                else:
                    xyz.append(np.array([float(split[1]), float(split[2]), float(split[3])]))
    return xyz

#print(psi4out2coords('cis1_bh267_m_vacumm_Cell_682.out'))

def RMSD_xyz(out1, out2, H = False):
    '''
    

    Parameters
    ----------
    out1 : str
        name of the first psi4out.
    out2 : str
        ame of the second psi4out.
    H : TYPE, bool
        if true use the H atoms for the calculation. The default is False.

    Returns
    -------
    float
        rmsd_kabsch with translate = True

    '''
    exyz1 = psi4out2coords(out1, element = True)
    exyz2 = psi4out2coords(out2, element = True)
    if H:
        exyz1 = np.array([item[1:] for item in exyz1], dtype=np.float)
        exyz2 = np.array([item[1:] for item in exyz2], dtype=np.float)
    else:    
        exyz1 = np.array([item[1:] for item in exyz1 if item[0] != 'H'], dtype=np.float)
        exyz2 = np.array([item[1:] for item in exyz2 if item[0] != 'H'], dtype=np.float)
    return float(kabsch_rmsd(exyz1, exyz2, translate = True))

def psi4_freq_check(out):
    '''
    

    Parameters
    ----------
    out : str
        name of the psi4 out file.

    Returns
    -------
    bool
        True if the freq calculation was made and any imaginary modes are present.
        False if not
    '''
    with open(out, 'rt', encoding='latin-1') as file:
        lines = file.readlines()
    cont = 0 #To check if the freq calcualtion was made
    for line in lines[::-1]: #The freq calculation is at the end, less line to process
        if 'Freq [cm^-1]' in line:
            cont += 1
            freqs = line.split()[2:]
            for f in freqs:
                '''
                 psi4 represent the imaginary number as 5i, 
                 if an error ocurreduring the float convertion, 
                 or it was converted but if less than cero (another error on sqrt) 
                 then return False
                '''
                try:
                    mt.sqrt(float(f))
                except:
                    return False
    if cont > 0:
        return True
    else:
        return False

def psi4_opt_check(out):
    '''
    

    Parameters
    ----------
    out : str
        name of the psi4 out file.

    Returns
    -------
    bool
        True if the opt was a succes.

    '''
    with open(out, 'rt', encoding='latin-1') as file:
        lines = file.readlines()
    for line in lines:
        if 'Optimization is complete!' in line: return True
    return False



def transpouse(array):
    '''
    

    Parameters
    ----------
    array : a list of list
        DESCRIPTION.

    Returns
    -------
    inv : numpy array
        the first row now is the first columns
        the second row now is the second columns
        etc.

    '''
    array = np.asarray(array)
    inv = np.zeros(len(array)*len(array[0]), dtype=array.dtype)
    inv.shape = (len(array[0]), len(array))
    for i in range(len(array)):
        for j in range(len(array[0])):
            inv[j,i]=array[i,j]
    return inv

def errors(dif):
    '''
    

    Parameters
    ----------
    dif : list of numbers (array)
        is a list that contain the diferences between the test set of values
        and the reference.
        It is crucial the order because the interpretation change
        test - reference is used in this function

    Returns
    -------
    RMSE : float
        Root Mean Square Error.
    MSE : float
        Mean Signed Error.
    MUE : float
        Mean Unsigned Error.

    '''
    RMSE = mt.sqrt(sum([item**2 for item in dif])/len(dif))
    MSE = sum(dif)/len(dif)
    MUE = sum([abs(item) for item in dif])/len(dif)
    return [RMSE, MSE, MUE]

def psi4_free_energy(out):
    with open(out, 'rt', encoding='latin-1') as file:
        lines = file.readlines()
    for line in lines[::-1]: #a[start:stop:step]
        if 'Total G, Free enthalpy at  298.15 [K]' in line:
            return float(line.split('[K]')[1].split()[0])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
===============================================================================
Created on    : 2020-2023
Author        : Alejandro Martínez León
Mail          : [alejandro.martinezleon@uni-saarland.de, ale94mleon@gmail.com]
Affiliation   : Jochen Hub's Biophysics Group
Affiliation   : Faculty of NS, University of Saarland, Saarbrücken, Germany
===============================================================================
DESCRIPTION   :
DEPENDENCIES  :
===============================================================================
"""
import os
from aleimi import templates, tools
def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]

def main(directory, elapse, launch = False):
    all_paths = []
    for root, dirs, files in os.walk(directory):
        root = os.path.abspath(root)
        if 'psi4.in' in files:
            if 'timer.dat' not in files:
                all_paths.append(root)

    for (i,elapsed_paths) in enumerate(chunks(all_paths, elapse)):
        T = templates.CONTINUE(elapsed_paths, machine = 'gwdg', name = f"elapse{i+1}")
        T.write(f"elapse{i+1}.sh")
        if launch == True:
            tools.run(f"sbatch elapse{i+1}.sh")
    

    
    



# -*- coding: utf-8 -*-
"""
===============================================================================
Created on    : Tue Dec  8 00:51:27 2020
Author        : Alejandro Martínez León
Mail          : [alejandro.marrtinezleon@uni-saarland.de, ale94mleon@gmail.com]
Affiliation   : Jochen Hub's Biophysics Group
Affiliation   : Faculty of NS, University of Saarland, Saarbrücken, Germany
===============================================================================
DESCRIPTION   :
DEPENDENCIES  :
===============================================================================
"""
# Se podria especificar los dos directorios en donde se desea comparar
# o un directorio de referencia (el de mayor nivel de teoria MP2)
# y luego todos los demas en los que se desee comparar con este, pues puedo tener
# B3LYP, HF y unos cuantos y entonces tendria que correr varias veces el mismo
# calcualo, seria recomendable que el directoria, especifique el nievel de teria 
# empleado para tomar esto como nombre.
# me queda implementar las funciones de los con signo y sin signo. Que lo que deberia es 
import calc_bond_angle as calc
import math as mt
import pandas as pd
from os import walk
import glob as glob

# =============================================================================
#      USER SPECIFICATIONS
# By defualt a single point energy is performed. You can specified if 
# freq is used or not.
# =============================================================================
higher_level_dir = ['MP2'] # reference
low_level_dir = ['HF']
check_function = calc.psi4_freq_check ## I can use psi4_opt_check insted of psi4_freq_check to compare the structures neve mind the imaginary modes

# When the list is used -1 need to be substrated. Because python array start at 0 not 1
bond_list = [(10,11),
             (11,12),
             (12,13),
             (13,14),
             (14,15),
             (15,10),
             (10,8),
             (8,9),
             (8,7),
             (7,5),
             (5,6),
             (5,2),
             (2,4),
             (2,3),
             (2,1),
             (1,16),
             (1,17),
             (1,18)
                     ]
angle_bond_list = [(11,10,8),
                   (10,8,9), 
                   (9,8,7),
                   (7,5,6),
                   (6,5,2),
                   (5,2,3),
                   (5,2,4),
                   (5,2,1),
                   (2,1,16),
                   (2,1,17),
                   (2,1,18)
                               ]
dihedral_list = [(11,10,8,7),
                  (10,8,7,5),
                  (8,7,5,2),
                  (7,5,6,19),
                  (7,5,2,1),
                  (5,2,1,16)
                             ]
# =============================================================================

py_bond_list = [list(map(lambda x:x-1, item)) for item in bond_list]
py_angle_bond_list = [list(map(lambda x:x-1, item)) for item in angle_bond_list]
py_dihedral_list = [list(map(lambda x:x-1, item)) for item in dihedral_list]



# Match each conformer/mols with the reference level
directories = higher_level_dir + low_level_dir
psi4_outs_alls = []
for directory in directories:
    psi4_outs_per_level = []
    for root, dirs, files in (walk('./'+directory)):    
        #print(root)
        #print(dirs)
        #print(files)
        outs = glob.glob(root+'/*.out')
        for out in outs:
            if 'myjob' not in out:
                psi4_outs_per_level.append(out)
    psi4_outs_alls.append(psi4_outs_per_level)

match = []
for item_higher in psi4_outs_alls[0]:
    match_temp = [item_higher]
    for lower_level in psi4_outs_alls[1:]:
        for item_lower in lower_level:
            if item_higher.split('/')[-1] == item_lower.split('/')[-1]:# check if the files have the same name
                match_temp.append(item_lower)
    if len(match_temp) > 1: match.append(match_temp) # if match something

# Each item in match contein a mol at different level of theories
# mol[0] has the higher level of theory, if there is some problem with the file
# delete the item of match
# is important to sorted in reverse for the pop function
idx_2_delete = sorted([i for i, mol in enumerate(match) if not check_function(mol[0])], reverse=True)

# Only save the level tested (the higher), the rest for these molecules will not tested
wrong_freq = [match.pop(idx)[0] for idx in idx_2_delete]



match_final = []
for mol in match:
    match_final_temp = [mol[0]]
    for lower_level in mol[1:]:
        if not check_function(lower_level):
            wrong_freq.append(lower_level)
        else:
            match_final_temp.append(lower_level)
    if len(match_final_temp) > 1: match_final.append(match_final_temp) # If more that only the higher level

with open('Wrong_freq.txt', 'wt') as file:
    file.write('The following files had freq or opt problems. For such molecules that the higher level of theory failed, it was not checked the rest level of theories.\n')
    [file.write(w+'\n') for w in wrong_freq]       



energy = []
RMSD_matrix = []
for mol in match_final:
    mol_energy = []
    mol_RMSD_matrix = []
    for level in mol:    
        mol_energy.append(calc.psi4_free_energy(level))
        mol_RMSD_matrix.append(calc.RMSD_xyz(level, mol[0]))
    energy.append(mol_energy)
    RMSD_matrix.append(mol_RMSD_matrix)

df_energy = pd.DataFrame(energy, index = [mol[0].split('/')[-1].split('.')[0] for mol in match_final], columns = [level.split('/')[1] for level in match_final[0]])
df_RMSD_matrix = pd.DataFrame(RMSD_matrix, index = [mol[0].split('/')[-1].split('.')[0] for mol in match_final], columns = [level.split('/')[1] for level in match_final[0]])
energy_RMSD_writer = pd.ExcelWriter('Energies_RMSD.ods')
df_energy.to_excel(energy_RMSD_writer, 'Energies')
df_RMSD_matrix.to_excel(energy_RMSD_writer, 'RMSD_xyz_ref_'+match_final[0][0].split('/')[1])
energy_RMSD_writer.save()


for i, mol in enumerate(match_final):   
    for j, level in enumerate(mol):
        xyz = calc.psi4out2coords(level)
        bonds = [calc.calc_bond_dist(*[xyz[item] for item in bond]) for bond in py_bond_list]
        angles = [calc.calc_angle_bond(*[xyz[item] for item in angle])*180/mt.pi for angle in py_angle_bond_list]
        dihedrals = [calc.calc_dihedral(*[xyz[item] for item in dihedral])*180/mt.pi for dihedral in py_dihedral_list]
        
        match_final[i][j] = [match_final[i][j], bonds, angles, dihedrals]
        


'''
Now, the structure of match_final is the following:
                                                            LevelS
    mol             MP2                                 HF                              B3LYP                       ...
    1    [name, bond, angle, dihedral]   [name, bond, angle, dihedral]          [name, bond, angle, dihedral]
    2    [name, bond, angle, dihedral]   [name, bond, angle, dihedral]          [name, bond, angle, dihedral]
    3    [name, bond, angle, dihedral]   [name, bond, angle, dihedral]          [name, bond, angle, dihedral]
    4   [name, bond, angle, dihedral]   [name, bond, angle, dihedral]          [name, bond, angle, dihedral]
    ...
    the next stepo want to reshape the data:
    for mol 1 
                                   Levels
    bonds    MP2                HF                                B3LYP
    1       x1                 x1                               x1
    2       x2                 x2                               x2
    3       x3                 x3                               x3
    4       x4                 x4                               x4
    ...
    and the same for the rest of the molecules. Each molecule will be exported to an excel file with three different sheets: bond, angles, dihedral
'''
for mol in match_final:
    columns = [item[0] for item in mol]
    # I need to rotate the data, this is done with transpouse ()
    df_bond = pd.DataFrame(calc.transpouse([item[1] for item in mol]), index = bond_list, columns = columns)
    df_angle = pd.DataFrame(calc.transpouse([item[2] for item in mol]), index = angle_bond_list, columns = columns)
    df_dihedral = pd.DataFrame(calc.transpouse([item[3] for item in mol]), index = dihedral_list, columns = columns)
    
    name = mol[0][0].split('/')[-1].split('.')[0]#+'benchmark'
    errors_bond = []
    errors_angle = []
    errors_dihedral = []
    writer = pd.ExcelWriter(name + '_benchmark.ods')
    for level in columns: #[1:] and them add a 0 in at the first position MP2/MP2

        dif_bond = list(df_bond[level] - df_bond[columns[0]]) # lower level - higher level(the higher level is constant for each molecule)
        dif_angle = list(df_angle[level] - df_angle[columns[0]])
        
        '''
        The following convention will be used.
        The dihedral angale is always in the interval -pi to pi.
        If both dihedrals are positive or negative at the same time:
            A negative difference means that the lower level move clockwise respect to the higher level
            And a positive difference means that the lower level move counterclockwise respect to the higher level.
        For mix signs:
        The correction down writed. The idea is take the same convention when both dihedrals are positive or negative at the same time
        This correction is needed because the real difference between 
        angels like -179 and 179 are not well estimated.
        The only differ in 2 degrees on the chemical structure.
        But -179-179 is not that.
        '''
        
        dif_dihedral = []
        for i in range(len(df_dihedral[level])): # Because I want to loop for each row
            lower_level = df_dihedral[level][i]
            higher_level = df_dihedral[columns[0]][i]
            if lower_level*higher_level < 0: # mix signs
                item1 = abs(higher_level) + abs(lower_level)
                item2 = 360 - item1
                if item1 < item2: # you need to get the small angle. Visualize the problem. Really important this check
                    if higher_level > 0:
                        item = -item1
                    else:
                        item = item1
                else:
                    if higher_level > 0:
                        item = item2
                    else:
                        item = -item2       
            else:
                item = lower_level - higher_level
            dif_dihedral.append(item)
        
        
        errors_bond.append(calc.errors(dif_bond))
        errors_angle.append(calc.errors(dif_angle))
        errors_dihedral.append(calc.errors(dif_dihedral))
    df_bond_errors = pd.DataFrame(calc.transpouse(errors_bond), index = ['RMSE', 'MSE', 'MUE'], columns=columns)
    df_angle_errors = pd.DataFrame(calc.transpouse(errors_angle), index = ['RMSE', 'MSE', 'MUE'], columns=columns)
    df_dihedral_errors = pd.DataFrame(calc.transpouse(errors_dihedral), index = ['RMSE', 'MSE', 'MUE'], columns=columns)
    
    
    df_bond_to_print = pd.concat([df_bond, df_bond_errors])
    df_angle_to_print = pd.concat([df_angle, df_angle_errors])
    df_dihedral_to_print = pd.concat([df_dihedral, df_dihedral_errors])

    df_bond_to_print.to_excel(writer, 'bond')
    df_angle_to_print.to_excel(writer, 'angle')
    df_dihedral_to_print.to_excel(writer, 'dihedral')
    writer.save()


# -*- coding: utf-8 -*-
"""
===============================================================================
Created on    : Sun Aug 25 18:49:11 2019
Author        : Alejandro Martínez León
Mail          : [amleon@instec.cu, ale94mleon@gmail.com]
Affiliation   : Chemical Systems Modeling Group,
Affiliation   : Faculty of Radiochemistry, InSTEC-University of Havana, Cuba.
===============================================================================
DESCRIPTION   : Usar con moderación. OpenBabel en dependencia del input pude dar discimiles errores.
DEPENDENCIES  : openbabel, glob
===============================================================================
"""


from openbabel import openbabel as ob
import glob as glob
import os, tempfile
from aleimi import tools



def obconvert(inpath, outpath, AddHydrogens = False, obabel = 'obabel'):
    """Convert  molecule ussing openbabel

    Args:
        input (str, path): input molecule.
        output (str, path): must have the extention of the molecule.
    """
    in_ext = os.path.basename(inpath).split('.')[-1]
    out_ext = os.path.basename(outpath).split('.')[-1] 

    if in_ext == 'pdbqt':
        tmpfile = tempfile.NamedTemporaryFile(suffix='.pdb')
        if AddHydrogens:
            tools.run(f"""
                    {obabel} {inpath} -O {tmpfile.name} -h > /dev/null 2>&1
                    {obabel} {tmpfile.name} -O {outpath} -h > /dev/null 2>&1
                    """)
        else:
            tools.run(f"""
                    {obabel} {inpath} -O {tmpfile.name} > /dev/null 2>&1
                    {obabel} {tmpfile.name} -O {outpath} > /dev/null 2>&1
                    """)
        tools.rm(tmpfile.name)


    else:
        obConversion = ob.OBConversion()
        obConversion.SetInAndOutFormats(in_ext, out_ext)
        mol = ob.OBMol()
        obConversion.ReadFile(mol, inpath)   # Open Babel will uncompressed automatically
        if AddHydrogens:
            mol.AddHydrogens()
        obConversion.WriteFile(mol, outpath)


if __name__ == '__main__':

    suppl = glob.glob('*')
    for i, item in enumerate(suppl):
        if '.py' in item:
            del suppl[i]
    for item in suppl:    
        obconvert(item, 'pdb')
